/*
 * ml3.c
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995-96)
 * this file belongs to the PUZZLE version 2.5.1 distribution
 *
 */


#define EXTERN extern
#define TRUE 1
#define FALSE 0

#define TSERR 0.5
#define YRERR 0.25
#define GAERR 0.25
#define LAERR 0.1

/* prototypes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "matrixut.h"
#include "ml.h"

/* abbreviation */
typedef unsigned long int ulint;

/* global variables */
int qa, qb, qc, qd;


/******************************************************************************/
/* chi square test (Numerical Recipes)                                        */
/******************************************************************************/

/* start code of Press et al. "Numerical Recipes in C" (1992) */

double gammln(double xx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}

#define ITMAX 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30

void gcf(double *gammcf, double a, double x, double *gln)
{
	int i;
	double an,b,c,d,del,h;

	*gln=gammln(a);
	b=x+1.0-a;
	c=1.0/FPMIN;
	d=1.0/b;
	h=d;
	for (i=1;i<=ITMAX;i++) {
		an = -i*(i-a);
		b += 2.0;
		d=an*d+b;
		if (fabs(d) < FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c) < FPMIN) c=FPMIN;
		d=1.0/d;
		del=d*c;
		h *= del;
		if (fabs(del-1.0) < EPS) break;
	}
	/* commented out by KSS and AvH */
	/* if (i > ITMAX) nrerror("a too large , ITMAX too small in gcf"); */
	
	*gammcf=exp(-x+a*log(x)-(*gln))*h;
}
#undef ITMAX
#undef EPS
#undef FPMIN

#define ITMAX 100
#define EPS 3.0e-7

void gser(double *gamser, double a, double x, double *gln)
{
	int n;
	double sum,del,ap;

	*gln=gammln(a);
	if (x <= 0.0) {
		/* commented out by KSS and AvH */
		/* if (x < 0.0) nrerror("x less than 0 in routine gser"); */
		*gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=1;n<=ITMAX;n++) {
			++ap;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS) {
				*gamser=sum*exp(-x+a*log(x)-(*gln));
				return;
			}
		}
		/* commented out by KSS and AvH */
		/* nrerror("a too large, ITMAX too small in routine gser"); */
		/* next line added by KSS and AvH */
		*gamser=sum*exp(-x+a*log(x)-(*gln));
		return;
	}
}
#undef ITMAX
#undef EPS

double gammq(double a, double x)
{
	double gamser,gammcf,gln;

	/* commented out by KSS and AvH */
	/* if (x < 0.0 || a <= 0.0)
		nrerror("Invalid arguments in routine gammq"); */
	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return 1.0-gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return gammcf;
	}
} /* end of NR code */

/* chi square test
   ef      expected frequencies (sum up to 1 !!)
   of      observed frequencies (sum up to the number of samples)
   numcat  number of categories
   returns critical significance level */
double chi2test(double *ef, int *of, int numcat)
{	
	double chi2, criticals, efn;
	int i, samples, below1, below5;
	
	chi2fail = FALSE;
	below1 = 0;
	below5 = 0;
	
	/* compute number of samples */
	samples = 0;
	for (i = 0; i < numcat; i++)
		samples = samples + of[i];
	
	/* compute chi square */
	chi2 = 0;
	for (i = 0; i < numcat; i++) {
		efn = ef[i]*((double) samples);
		if (efn < 1.0) below1++;
		if (efn < 5.0) below5++;
		chi2 = chi2 + ((double) of[i]-efn)*((double) of[i]-efn)/efn;
	}

	/* compute significance if chi2 were the critical value */
	criticals = gammq((numcat - 1)*0.5, chi2*0.5);
	
	/* no expected frequency category (sum up to # samples) below 1.0 */
	if (below1 > 0) chi2fail = TRUE;
	/* no more than 1/5 of the frequency categories below 5.0 */
	if (below5 > (int) floor(samples/5.0)) chi2fail = TRUE;

	return criticals;
}

/* compare general base frequencies with frequencies of taxon i with chi square */
double homogentest(int taxon)
{
	return chi2test(Freqtpm, Basecomp[taxon], gettpmradix());
}


/******************************************************************************/
/* discrete Gamma-distribution                                                */
/******************************************************************************/

/* inverse of regularized incomplete gamma function
   -  algorithm adopted from the Mathematica package
      "InverseStatisticalFunctions.m" by Jerry B. Keiper
   -  copyright 1989-1993, Wolfram Research, Inc.
   -  the original MMA code can be freely obtained from the
      web site http://www.wri.com/mathsource
   -  note that the original MMA code is *much* more
      sophisticated with regard to the accuracy of the result
   restrictions: a > 0  &&  x >= 0  &&  1 > pp >= 0 */
double invgamreg(double a, double x, double pp)
{
#define EPS 0.000001
#define REPEATS 20

	int i, cflag;
	double gamx, p, gama, x0, x1, x2, xold, temp, pa, dy;
	 

	if (x > a) cflag = TRUE;
	else cflag = FALSE;

	if (cflag) {
		gamx = gammq(a,x);
		p = pp + 1.0 - gamx;
	} else {
		gamx = (gammq(a,0.0)-gammq(a,x));
		p = pp + gamx;
	}

	if (p == 0.0) return 0.0;
	if(a == 1.0) {
		if (cflag) p = gamx - pp;
		else p = 1.0 - p;
		return (-log(p));
	}

	gama = exp(gammln(a));
	pa = pow((p*a*gama),1.0/a);
	temp = pa/(a+1.0);

	x0 = pa*(1.0+temp*(1.0+temp*((5.0+3.0*a)/(2.0*(a+2.0))+temp*((31.0+a*
		(33.0+8.0*a))/(3.0*(2.0+a)*(3.0+a))+temp*((2888.0+a*(5661.0+a*
		(3971.0+a*(1179.0+125.0*a))))/(24.0*(2.0+a)*(2.0+a)*(3.0+a)*(4.0+a))+
		temp*((9074.0+a*(20997.0+a*(18375.0+a*(7575.0+a*(1471.0+108.0*a)))))/
		(10.0*(2.0+a)*(2.0+a)*(3.0+a)*(4.0+a)*(5.0+a))))))));

	if(x0 > a) {
		if (cflag)
			p = gamx - pp;
		else {
			cflag = TRUE;
			p = 1.0 - p;
		}
	} else cflag = FALSE;
	
	if (p < 0.0) {
		printf("\nERROR:  Inverse of incomplete gamma function\n");
	}

	/* now use Taylor series to quartically converge to x0 */
	xold = x0;
	p *= gama;
	
	for (i = 0; i < REPEATS; i++) {
		if (cflag)
			dy = p-gama*gammq(a,x0);
		else
			dy =  p-gama*(gammq(a,0.0) - gammq(a,x0));

		if (fabs(dy) < EPS) return x0;	
		if (cflag) dy = -dy;

		x1 = exp(x0)*dy*pow(x0,-a);
		x2 = x1*((1.0-a+x0)/2.0+x1*((1.0+a*(-3.0+2.0*a)+(4.0-4.0*a+2.0*x0)*x0)/
			6.0+x1*(1.0+a*(-6.0+a*(11.0-6.0*a))+x0*(11.0+a*(-29.0+18.0*a)+
			x0*(18.0-18.0*a+6.0*x0)))/24.0));
		x1 *= x0*(1.0+x2);
		x0 += x1;
	}
	/* no convergence ! */
	return xold;
}
#undef EPS
#undef REPEATS

/* compute rates of each category when rates are Gamma-distributed */
void updaterates()
{
	int i;
	double mean;
	
	if (numcats == 1) {
		Rates[0] = 1.0;
		return;
	}
	Rates[0] = invgamreg(Galpha,0.0,1.0/(2.0*numcats))/Galpha;
	for (i = 1; i < numcats; i++) {
		Rates[i] = invgamreg(Galpha,0.0,(2.0*i+1.0)/(2.0*numcats))/Galpha;
		/* in case that invgamreg did not converge */
		if (Rates[i] < Rates[i-1])
			Rates[i] = Rates[i-1];
	}
	/* make mean of discrete distribution equal to one */
	mean = 0.0;
	for (i = 0; i < numcats; i++) mean += Rates[i];
	mean = mean/(double)numcats;
	for (i = 0; i < numcats; i++) Rates[i] = Rates[i]/mean;
	/* check for very small rates */
	for (i = 0; i < numcats; i++)
		if (Rates[i] < 0.000001) Rates[i] = 0.000001;
}

/******************************************************************************/
/* minimization of a function by Brents method (Numerical Recipes)            */
/******************************************************************************/

#define ITMAX 100
#define CGOLD 0.3819660
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define ZEPS 1.0e-10
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

/* Brents method in one dimension */
double brent(double ax, double bx, double cx, double (*f)(double), double tol,
	double *foptx, double *f2optx, double fax, double fbx, double fcx)
{
	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double xw,wv,vx;
	double e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=bx;
	fx=fbx;
	if (fax < fcx) {
		w=ax;
		fw=fax;
		v=cx;
		fv=fcx;
	} else {
		w=cx;
		fw=fcx;
		v=ax;
		fv=fax;	
	}
	for (iter=1;iter<=ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*foptx = fx;
			xw = x-w;
			wv = w-v;
			vx = v-x;
			*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
				(v*v*xw + x*x*wv + w*w*vx);
			return x;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=(*f)(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	*foptx = fx;
	xw = x-w;
	wv = w-v;
	vx = v-x;
	*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
		(v*v*xw + x*x*wv + w*w*vx);
	return x;
}
#undef ITMAX
#undef CGOLD
#undef ZEPS
#undef SHFT
#undef SIGN
#undef GOLD
#undef GLIMIT
#undef TINY

/* one-dimensional minimization - as input a lower and an upper limit and a trial
  value for the minimum is needed: xmin < xguess < xmax
  the function and a fractional tolerance has to be specified
  onedimenmin returns the optimal x value and the value of the function
  and its second derivative at this point
  */
double onedimenmin(double xmin, double xguess, double xmax, double (*f)(double),
	double tol, double *fx, double *f2x)
{
	double eps, optx, ax, bx, cx, fa, fb, fc;
		
	/* first attempt to bracketize minimum */
	eps = xguess*tol*50.0;
	ax = xguess - eps;
	if (ax < xmin) ax = xmin;
	bx = xguess;
	cx = xguess + eps;
	if (cx > xmax) cx = xmax;
	
	/* check if this works */
	fa = (*f)(ax);
	fb = (*f)(bx);
	fc = (*f)(cx);

	/* if it works use these borders else be conservative */
	if ((fa < fb) || (fc < fb)) {
		if (ax != xmin) fa = (*f)(xmin);
		if (cx != xmax) fc = (*f)(xmax);
		optx = brent(xmin, xguess, xmax, f, tol, fx, f2x, fa, fb, fc);
	} else
		optx = brent(ax, bx, cx, f, tol, fx, f2x, fa, fb, fc);

	return optx; /* return optimal x */
}

/******************************************************************************/
/* parameter estimation                                                       */
/******************************************************************************/

/* compute sample mean and standard deviation of sample mean */
void computestat(double *data, int n, double *mean, double *err)
{	
	int i;
	double sum;
	
	sum = 0;
	for (i = 0; i < n; i++) sum += data[i];
	(*mean) = sum/(double) n;
	
	sum = 0;
	for (i = 0; i < n; i++) sum += (data[i] - (*mean))*(data[i] - (*mean));
	if (n != 1)
		(*err) = sqrt(sum/((double) n*(n-1))); /* unbiased estimator */
	else
		(*err) = 0.0; /* if n == 1 */
}

/* compute ML value of quartet (a,b,c,d) */
double quartetml(int a, int b, int c, int d)
{
	double d1, d2, d3;

	/* compute approximate ML for all topologies */
	d1 = quartet_alklhd(a,b,c,d); /* (a,b)-(c,d) */
	d2 = quartet_alklhd(a,c,b,d); /* (a,c)-(b,d) */
	d3 = quartet_alklhd(a,d,b,c); /* (a,d)-(b,c) */
	
	/* looking for max(d1, d2, d3) */						
	if (d1 < d2) { /* d2 > d1 */
		if (d2 < d3) { /* d3 > d2 > d1 */
			/* d3 maximum */ 
			return d3;
		} else { /* d2 >= d3 > d1 */
			/* d2 maximum */
			return d2;
		}
	} else { /* d1 >= d2 */
		if (d1 < d3) { /* d3 > d1 >= d2 */
			/* d3 maximum */
			return d3;
		} else { /* d1 >= d2 && d1 >= d3 */
			/* d1 maximum */
			return d1;
		}
	}
}

/* optimization function AlphaBeta */
double optab(double x)
{
	if (x < MINAB) AlphaBeta = MINAB;
	else if (x > MAXAB) AlphaBeta = MAXAB;
	else AlphaBeta = x;
	tranprobmat();
	distupdate(qa, qb, qc, qd);	
	return (-quartetml(qa, qb, qc, qd));
}

/* optimization function AlphaYR */
double optyr(double x)
{
	if (x < MINYR) AlphaYR = MINYR;
	else if (x > MAXYR) AlphaYR = MAXYR;
	else AlphaYR = x;
	tranprobmat();
	distupdate(qa, qb, qc, qd);
	return (-quartetml(qa, qb, qc, qd));
}

/* optimization function Galpha */
double optga(double x)
{
	if (x < MINGA) Galpha = MINGA;
	else if (x > MAXGA) Galpha = MAXGA;
	else Galpha = x;
	updaterates();
	distupdate(qa, qb, qc, qd);
	return (-quartetml(qa, qb, qc, qd));
}

/* optimization function lambda */
double optla(double x)
{
	if (x < MINLA) lambda = MINLA;
	else if (x > MAXLA) lambda = MAXLA;
	else lambda = x;
	distupdate(qa, qb, qc, qd);
	return (-quartetml(qa, qb, qc, qd));
}

#define EPS1 0.01  /* for model parameters */
#define EPS2 0.01  /* for rate variation parameters */
#define MAXITS 5   /* maximum number of iterations */

/* optimize quartet a, b, c, d */
/* optimizing along the basis vectors may take a very long time ... */
void optimquartet(int a, int b, int c, int d)
{
	int it, nump, change;
	double ABold, YRold, GAold, LAold;
	double fx, f2x;

	qa = a;
	qb = b;
	qc = c;
	qd = d;

	it = 0;
	nump = 0;

	/* initialize start values with average value */
	if ((SH_optn || nuc_optn) && optim_optn && (data_optn == 0)) {
		AlphaBeta = tsmean*2.0;
		nump++;
	}
	if ((nuc_optn && TN_optn) && optim_optn && (data_optn == 0)) {
		AlphaYR = yrmean;
		nump++;
	}
	if (rate_optim && numcats > 1) {
		Galpha = gamean;
		nump++;
	}
	if (lambda_optim && numcats > 1) {
		lambda = lamean;
		nump++;
	}	

	do { /* repeat until nothing changes any more */
		it++;
		change = FALSE;

		/* optimize Ts/Tv ratio */
		if ((SH_optn || nuc_optn) && optim_optn && (data_optn == 0)) {
		
			if (AlphaBeta <= MINAB) AlphaBeta = MINAB+0.1;
			if (AlphaBeta >= MAXAB) AlphaBeta = MAXAB-0.1;
			ABold = AlphaBeta;
			AlphaBeta = onedimenmin(MINAB, AlphaBeta, MAXAB, optab, EPS1, &fx, &f2x);
			if (AlphaBeta < MINAB) AlphaBeta = MINAB;
			if (AlphaBeta > MAXAB) AlphaBeta = MAXAB;
			if (fabs(AlphaBeta - ABold) > 2*TSERR) change = TRUE;
		
		}

		/* optimize AlphaYR */
		if ((nuc_optn && TN_optn) && optim_optn && (data_optn == 0)) {

			if (AlphaYR <= MINYR) AlphaYR = MINYR+0.01;
			if (AlphaYR >= MAXYR) AlphaYR = MAXYR-0.01;
			YRold = AlphaYR;
			AlphaYR = onedimenmin(MINYR, AlphaYR, MAXYR, optyr, EPS1, &fx, &f2x);
			if (AlphaYR < MINYR) AlphaYR = MINYR;
			if (AlphaYR > MAXYR) AlphaYR = MAXYR;
			if (fabs(AlphaYR - YRold) > YRERR) change = TRUE;

		}
			
		/* optimize Galpha */
		if (rate_optim && numcats > 1) {
		
			if (Galpha <= MINGA) Galpha = MINGA+0.01;
			if (Galpha >= MAXGA) Galpha = MAXGA-0.01;
			GAold = Galpha;
			Galpha = onedimenmin(MINGA, Galpha, MAXGA, optga, EPS2, &fx, &f2x);
			if (Galpha < MINGA) Galpha = MINGA;
			if (Galpha > MAXGA) Galpha = MAXGA;
			if (fabs(Galpha - GAold) > GAERR) change = TRUE;
		
		}
			
		/* optimize lambda */
		if (lambda_optim && numcats > 1) {

			if (lambda <= MINLA) lambda = MINLA+0.01;
			if (lambda >= MAXLA) lambda = MAXLA-0.01;
			LAold = lambda;
			lambda = onedimenmin(MINLA, lambda, MAXLA, optla, EPS2, &fx, &f2x);
			if (lambda < MINLA) lambda = MINLA;
			if (lambda > MAXLA) lambda = MAXLA;
			if (fabs(lambda - LAold) > LAERR) change = TRUE;

		}

		if (nump == 1) return;
		
	} while (it != MAXITS && change);

	return;
}
#undef EPS1
#undef EPS2
#undef MAXITS

#define MINPERTAXUM 2
#define MAXPERTAXUM 6

/* estimate parameters by ML maximizing random quartets */
void optimparams()
{
	double tsmeanold, yrmeanold, gameanold, lameanold;
	dvector tstvlist, abyrlist, galphalist, lalist;
	int i, j, k, ID, fin;
	ivector isfree, taxon;
	ulint minqts, maxqts, n;

	
	isfree = new_ivector(Maxspc);
	taxon = new_ivector(4);

	/* number of quartets to be investigated */
	minqts = (ulint) floor(0.25 * MINPERTAXUM * Maxspc) + 1;
	maxqts = (ulint) floor(0.25 * MAXPERTAXUM * Maxspc) + 1;
	if (Maxspc == 4) {
		minqts = (ulint) 1;
		maxqts = (ulint) 1;
	}
	
	tstvlist = new_dvector(maxqts);
	abyrlist = new_dvector(maxqts);
	galphalist = new_dvector(maxqts);
	lalist = new_dvector(maxqts);
	
	/* initialize averages */
	tsmean = AlphaBeta/2.0;
	yrmean = AlphaYR;
	gamean = Galpha;
	lamean = lambda;

	fin = FALSE;
		
	/* investigate maxqts random quartets */
	for (n = 0; n < maxqts; n++) {
	
		/* choose random quartet */		
		for (k = 0; k < Maxspc; k++) isfree[k] = TRUE;
			for (k = 0; k < 4; k++) {
				/* random integer in the range [0, Maxspc-1-k] */
				ID = randominteger(Maxspc - k);
				i = -1;
				j = -1;
				do {
					i++;
					if (isfree[i] == TRUE) j++; 
				} while ( j != ID);
				taxon[k] = i;
				isfree[i] = FALSE;
			}

		printf("\n        Optimizing random quartet # %ld ", n+1);
		printf("(%ld to %ld quartets necessary):\n",
			minqts, maxqts);

		/* optimize parameters on current quartet */
		optimquartet(taxon[0], taxon[1], taxon[2], taxon[3]);

		tsmeanold = tsmean;
		yrmeanold = yrmean;
		gameanold = gamean;
		lameanold = lamean;
		tstvlist[n] = AlphaBeta/2.0;
		abyrlist[n] = AlphaYR;
		galphalist[n] = Galpha;
		lalist[n] = lambda;
		computestat(tstvlist, n+1 , &tsmean, &tserr);
		computestat(abyrlist, n+1 , &yrmean, &yrerr);
		computestat(galphalist, n+1 , &gamean, &gaerr);
		computestat(lalist, n+1 , &lamean, &laerr);

		if ((SH_optn || nuc_optn) && optim_optn && (data_optn == 0))
			printf("          Transition/transversion parameter: %.2f\n",
				tstvlist[n]);
		if ((nuc_optn && TN_optn) && optim_optn && (data_optn == 0))
			printf("          Y/R transition parameter: %.2f\n",
				abyrlist[n]);
		if (rate_optim && numcats > 1)
			printf("          Rate heterogenity parameter: %.2f\n",
				galphalist[n]);
		if (lambda_optim && numcats > 1)
			printf("          Rate correlation parameter: %.2f\n",
				lalist[n]);

		/* check whether the means are converging */
		if (n > minqts-2) {
			if ((fabs(tsmean-tsmeanold) < TSERR) &&
				(fabs(yrmean-yrmeanold) < YRERR) &&
				(fabs(gamean-gameanold) < GAERR) &&
				(fabs(lamean-lameanold) < LAERR))
					fin = TRUE;
		}

		/* investigate at least minqts quartets */
		if (n > minqts-2 && (fin || n > maxqts-2)) break;
	}

	/* round estimated numbers to 2 digits after the decimal point */
	if (tserr != 0.0) tsmean = floor(100.0*tsmean+0.5)/100.0;
	if (yrerr != 0.0) yrmean = floor(100.0*yrmean+0.5)/100.0;
	if (gaerr != 0.0) gamean = floor(100.0*gamean+0.5)/100.0;
	if (laerr != 0.0) lamean = floor(100.0*lamean+0.5)/100.0;

	/* update ML engine */
	AlphaBeta = 2.0*tsmean;
	AlphaYR = yrmean;
	Galpha = gamean;
	lambda = lamean;
	updaterates();
	tranprobmat();

	free_ivector(isfree);
	free_ivector(taxon);
}

#undef MINPERTAXUM
#undef MAXPERTAXUM
